from os import getenv
from .stv import STV
from .stv_progress import STVProgress

VOTES_LIMIT = int(getenv('VOTES_LIMIT', 50))


def pos_to_json(pos, initquota, winners_quota, viewvoter):
    j = {
        'round': pos.round,
        'subround': pos.subround,
        'loopcount': pos.loopcount,
        'looptype': pos.looptype,
        'message': pos.message,
        'candidates': {},
        'viewballot': None,
        'waste': round(sum(pos.waste.values()), 2)
    }
    for status, candlist in [('winner', pos.winners), ('active', pos.active), ('deactivated', pos.deactivated),
                             ('excluded', pos.excluded)]:
        for cand in candlist:
            quota = round(winners_quota[cand.code] if status == 'winner' else initquota, 2)
            j['candidates'][cand.code] = {'votes': round(cand.votes, 2), 'status': status, 'quota': quota}

    if viewvoter is not None:
        j['viewballot'] = []
        for vf in pos.votefractions.values():
            if vf.voterid == viewvoter:
                ballotline = {'ccode': vf.candidatecode, 'fraction': round(vf.fraction, 2), 'status': vf.status[0]}
                j['viewballot'].append(ballotline)

    return j


def lambda_handler(event, context):
    # Preliminary checks
    usegroups = event['usegroups']
    reactivation = event['reactivation']
    groups = event['groups']
    candidates = event['candidates']
    votes = event['votes']
    viewvoter = event.get('viewvoter')

    if len(votes) > VOTES_LIMIT:
        return get_error('Function', 'limit is {} votes'.format(VOTES_LIMIT))

    stv = STV(usegroups, reactivation)

    for group in groups:
        stv.add_group(group['name'], group['seats'])

    for candidate in candidates:
        stv.add_candidate(candidate['code'], candidate['name'], candidate['group'])

    for vote in votes:
        stv.add_voter(vote['voterid'], vote['ballot'])

    if viewvoter not in stv.voters:
        viewvoter = None
    stvp = STVProgress(stv)
    # Get Quotas
    initquota = stv.quota
    winners_quota = {cand.code: cand.wonatquota for cand in stv.winners}

    loops = []
    for t, pos in stvp.get_tansform_and_position():
        loops.append(pos_to_json(pos, initquota, winners_quota, viewvoter))

    # Create links
    lastroundli = len(loops) - 1
    lastsubroundli = len(loops) - 1
    lastround = loops[lastroundli]['round']
    lastsubround = loops[lastsubroundli]['subround']

    for i, loop in list(enumerate(loops))[::-1]:
        loop['nextRound'] = lastroundli
        loop['nextSubround'] = lastsubroundli

        if loop['round'] != lastround:
            lastround = loop['round']
            lastroundli = i

        if loop['round'] != lastround or loop['subround'] != lastsubround:
            lastsubround = loop['subround']
            lastsubroundli = i

    lastroundli = 0
    lastsubroundli = 0
    lastround = loops[lastroundli]['nextRound']
    lastsubround = loops[lastsubroundli]['nextSubround']

    for i, loop in enumerate(loops):
        loop['previousRound'] = lastroundli
        loop['previousSubround'] = lastsubroundli

        if loop['nextRound'] != lastround:
            lastround = loop['nextRound']
            lastroundli = i

        if loop['nextRound'] != lastround or loop['nextSubround'] != lastsubround:
            lastsubround = loop['nextSubround']
            lastsubroundli = i

    return {'quota': stv.quota, 'loops': loops, 'viewvoter': viewvoter}


def get_error(errortype, msg):
    return {'errorType': errortype, 'errorMessage': msg}



from typing import List, Optional, Generator
from collections import namedtuple
from .stv import STV

Candidate = namedtuple('Candidate', ['code', 'votes'])
VoteFraction = namedtuple('VoteFraction', ['voterid', 'fraction', 'candidatecode', 'status'])


class Position:
    # Loop Type
    UNKNOWN = 0
    REDUCTION = 1
    ALLOCATION = 2
    LOSS = 3
    WIN = 4

    def __init__(self, stv: STV, status, previous_position):
        def tupelize_candidate_list(candlist):
            return [Candidate(c.code, c.votes) for c in candlist]

        self.round = stv.rounds
        self.subround = stv.subrounds
        self.loopcount = stv.loopcount

        if status.winner is not None:
            self.looptype = self.WIN
            self.message = 'Win: ' + status.winner.name
        elif status.loser is not None:
            self.looptype = self.LOSS
            self.message = 'Loss: ' + status.loser.name
        elif stv.reductioncount > 0:
            self.looptype = self.REDUCTION
            self.message = 'Reductions: {}'.format(stv.reductioncount)
        elif stv.allocationcount > 0:
            self.looptype = self.ALLOCATION
            self.message = 'Allocations: {}'.format(stv.allocationcount)
        else:
            self.looptype = self.UNKNOWN
            self.message = 'Beginning'

        self.excluded_group = status.excluded_by_group[0].group.name if status.excluded_by_group else None
        if self.excluded_group:
            self.message += '\nExclusion of group: ' + self.excluded_group

        self.winners = tupelize_candidate_list(stv.winners)
        self.active = tupelize_candidate_list(stv.active)
        self.deactivated = tupelize_candidate_list(stv.deactivated)
        self.excluded = tupelize_candidate_list(stv.excluded)

        self.votefractions = {}
        self.waste = {}  # Key is VoterID
        for voter in stv.voters.values():
            vid = voter.uid
            self.waste[vid] = voter.waste
            for vl in voter.votelinks:
                ccode = vl.candidate.code
                vlstatuscode = {vl.EXCLUDED: 'Excluded', vl.DEACTIVATED: 'Deactivated', vl.ACTIVE: 'Active',
                                vl.PARTIAL: 'Partial', vl.FULL: 'Full'}[vl.status]
                self.votefractions[(vid, ccode)] = VoteFraction(vid, vl.weight, ccode, vlstatuscode)

        self.nexttransform: Optional[Transform] = None

        if previous_position is not None:
            self.add_previous(previous_position)

    @property
    def hasdecision(self) -> bool:
        return self.looptype >= self.LOSS

    def add_previous(self, previous: 'Position') -> None:
        previous.nexttransform = t = Transform(self)

        for k, nvf in self.votefractions.items():
            t.add_difference(previous.votefractions[k], nvf)


class Transform:
    def __init__(self, nextposition: Position):
        self.nextposition = nextposition
        self.returnvfs: List[VoteFraction] = []
        self.sendvfs: List[VoteFraction] = []

    def add_difference(self, previousvf: VoteFraction, nextvf: VoteFraction) -> None:
        weightdiff = nextvf.fraction - previousvf.fraction
        if weightdiff != 0:
            vflist = self.sendvfs if weightdiff > 0 else self.returnvfs
            vflist.append(VoteFraction(nextvf.voterid, abs(weightdiff), nextvf.candidatecode, nextvf.status))


class STVProgress:
    def __init__(self, stv: STV):
        """ receives a fresh stv instance and creates all positions and transforms """
        self.startpos = None
        currentpos = None
        for status in stv.start():
            if status.yieldlevel >= 0:
                newpos = Position(stv, status, currentpos)

                if self.startpos is None:
                    self.startpos = newpos

                currentpos = newpos

    def get_tansform_and_position(self) -> Generator:
        if self.startpos is None:
            raise Exception("STV Progress could not initialize")
        yield None, self.startpos
        t = self.startpos.nexttransform
        while t is not None:
            yield t, t.nextposition
            t = t.nextposition.nexttransform



import argparse
import sys
from importlib import resources
from .stv import STV, STVStatus, STVSetupException


def main() -> None:
    parser = argparse.ArgumentParser(prog='stvlebanon', description="Command-Line interface to STV Lebanon")
    parser.add_argument('-g', dest='group', action='store_true', help="Use group quotas")
    parser.add_argument('-n', dest='reactivation', action='store_false', help="No reactivation")
    parser.add_argument('-l', dest='level', type=int, default=0, help="Level of monitoring. 1=Round 2=Subround 3=Loop")
    parser.add_argument('-w', dest='watch', default="", metavar="VOTERID", help="View Voter Situation at every round")
    parser.add_argument('-s', dest='sample', action='store_true', help="Load sample data")
    parser_result = parser.parse_args()

    use_groups: bool = parser_result.group
    reactivation: bool = parser_result.reactivation
    viewlevel: int = min(max(parser_result.level, 0), 3) + 1  # So it matches STVStatus levels
    viewvoter: str = parser_result.watch
    load_samples: bool = parser_result.sample

    print("Use -h to see running options\n")
    print("Groups:", use_groups)
    print("Reactivation:", reactivation)
    print("View Level:", {STVStatus.END: "Result", STVStatus.ROUND: "Round", STVStatus.SUBROUND: "Subround",
                          STVStatus.LOOP: "Loop"}[viewlevel])
    print("Watching:", viewvoter or "<None>")

    stv = setup(use_groups, reactivation, load_samples)

    if viewvoter and viewvoter not in stv.voters:
        print(f"\nWarning: Could not find Voter with ID: {viewvoter}")
        viewvoter = ""

    print(f"\nSeats: {stv.totalseats}\nTotal Votes: {len(stv.voters)}  Quota: {formatvote(stv.quota)}\n")
    
    for status in stv.start():
        if status.yieldlevel <= viewlevel and status.yieldlevel != status.BEGIN:
            if status.yieldlevel == status.INITIAL:
                print("Initial Round\n")
            elif viewlevel >= status.ROUND:
                print("Round:", '.'.join(map(str, [stv.rounds, stv.subrounds, stv.loopcount][:viewlevel-status.END])))

                if status.winner is not None:
                    print("Win:", status.winner.name)
                elif status.loser is not None:
                    print("Loss:", status.loser.name)
                elif stv.allocationcount > 0:
                    print("Allocations:", stv.allocationcount)
                elif stv.reductioncount > 0:
                    print("Reductions:", stv.reductioncount)
                print()

            if status.excluded_by_group:
                print("The following candidates have been excluded because their group quota has been met:")
                for c in status.excluded_by_group:
                    print(c.name)
                print()

            if status.reactivated:
                print("The following candidates have been returned to the active list:")
                for c in status.reactivated:
                    print(c.name)
                print()

            print_lists(stv, viewvoter)

            print("---------------------------\n")
            if not status.yieldlevel == status.END:
                try:
                    input("Press any key to continue to next round...")
                except KeyboardInterrupt:
                    print("\n\nExiting program...\n")
                    sys.exit()
                print()
            else:
                print("Votes Finished")
                for group in stv.groups.values():
                    print(group.name, group.seatswon, '/', group.seats)
                print("Waste Percentage:", formatratio(stv.totalwaste / len(stv.voters)))


def setup(usegroups: bool, reactivationmode: bool, load_samples: bool = False) -> STV:
    """ Import from local files, create and return STV instance """

    def local_or_sample(filename: str) -> str:
        if load_samples:
            with resources.path('stv_lebanon.samples', filename) as new_filename:
                filename = new_filename
            print("Using sample:", filename)
        return filename

    stv = STV(usegroups, reactivationmode)
    try:
        # Fill Objects
        with open(local_or_sample('Groups.csv'), 'r') as f:
            for i, line in enumerate(f, start=1):
                line = line.strip()  # Skip empty lines
                try:
                    groupname, seats = line.split(',')
                    stv.add_group(groupname, int(seats))
                except ValueError:
                    raise STVSetupException(f"Could not decode group at line {i}")

        with open(local_or_sample('Candidates.csv'), 'r') as f:
            for i, line in enumerate(f, start=1):
                line = line.strip()
                if line:  # Skip empty lines
                    try:
                        uid, name, groupname = line.split(',')
                        stv.add_candidate(uid, name, groupname)
                    except ValueError:
                        raise STVSetupException(f"Could not decode candidate at line {i}")

        with open(local_or_sample('Votes.csv'), 'r') as f:
            for i, line in enumerate(f, start=1):
                line = line.strip()
                if line:  # Skip empty lines
                    try:
                        uid, *ballot = line.split(',')
                        stv.add_voter(uid, ballot)
                    except ValueError:
                        print("Setup Warning: Could not decode voter at line", i)

    except (FileNotFoundError, STVSetupException) as e:
        if isinstance(e, FileNotFoundError):
            print(f"\nError: Missing file '{e.filename}'. Please make sure you have the following 3 files in your "
                  "current directory: 'Groups.csv', 'Candidates.csv', 'Votes.csv'")
        else:
            print("\nSetup Error:", e)
        sys.exit(1)
    return stv


def print_lists(stv: STV, voterid: str) -> None:
    """ Prints candidate list and optionally, voter's ballot """
    statcands = [('W', stv.winners), ('A', stv.active), ('D', stv.deactivated[::-1]), ('E', stv.excluded[::-1])]

    for status, candidates in statcands:
        for candidate in candidates:
            print(formatname(candidate.name) + status, formatvote(candidate.votes))
    print('------------')
    print(formatname('Total Waste') + ' ', formatvote(stv.totalwaste))

    if voterid:  # If not empty string. Already checked after setup
        voter = stv.voters[voterid]
        print('\n' + voter.uid, 'list:')
        for vl in voter.votelinks:
            statusdescription = {vl.EXCLUDED: 'Excluded', vl.DEACTIVATED: 'Deactivated', vl.ACTIVE: 'Active',
                                 vl.PARTIAL: 'Partial', vl.FULL: 'Full'}[vl.status]
            print(formatname(vl.candidate.name), formatratio(vl.weight), ' ' + statusdescription)
        print(formatname('Waste'), formatratio(voter.waste))


def formatname(v):
    return '{:<20}'.format(v)


def formatvote(v):
    return '{:,.2f}'.format(v)


def formatratio(v):
    return '{:>5.0%}'.format(v)



from typing import List, Dict, Generator, Final, Optional


class STVSetupException(Exception):
    pass


class Group:
    def __init__(self, name: str, seats: int):
        self.name = name
        self.seats = seats
        self.seatswon = 0

    @property
    def is_full(self) -> bool:
        return self.seatswon >= self.seats


class Candidate:
    def __init__(self, code: str, name: str, group: Group):
        self.code = code
        self.name = name
        self.group = group
        self.votelinks: List[VoteLink] = []

        # Running Attributes
        self._votes: float = 0
        self.dorefreshvotes = True
        self.wonatquota: float = 0
        self.doreduction = False

    def __repr__(self):
        return f"Candidate({self.code}, {self.name})"

    @property
    def votes(self) -> float:
        """ Cache result to improve performance """
        if self.dorefreshvotes:
            self.dorefreshvotes = False
            self._votes = 0
            for vl in self.votelinks:
                self._votes += vl.weight
        return self._votes

    def reduce(self) -> None:
        """
        Now that the candidate has surplus votes,
        see who from the partial supporters can become full supporters,
        then reduce the weight taken from all full supporters
        """
        self.doreduction = False

        # Create ordered list of partial votelinks from lowest to highest and list of full votelinks
        partialvls = []
        fullvls = []
        for vl in self.votelinks:
            if vl.status == vl.FULL:
                fullvls.append(vl)
            elif vl.status == vl.PARTIAL and vl.weight > 0:
                partialvls.append(vl)
        partialvls.sort(key=lambda x: x.weight)

        totalsupporters = len(fullvls) + len(partialvls)
        partialcount = 0
        partialweight: float = 0
        for vl in partialvls + fullvls:
            threshold = (self.wonatquota - partialweight) / (totalsupporters - partialcount)
            # Threshold is the weight at which a VoteLink can qualify as FULL and the weight at which FULL support
            # will be reduced.
            # As loop progresses, the threshold will increase than stabilize when supporters are able to
            # become full supporters
            if vl.status == vl.PARTIAL:
                if vl.weight < threshold:
                    # Phase 1. partial supporters who cannot give full support
                    partialcount += 1
                    partialweight += vl.weight
                    # This will increase  the threshold on next iteration
                else:
                    # Phase 2. partial supporter can now fully support candidate
                    vl.status = vl.FULL

            # Reduce full support weight for old and new full supporters
            if vl.status == vl.FULL:
                vl.weight = threshold
                self.dorefreshvotes = True  # Have to recalculate weight
                vl.voter.doallocate = True
                vl.voter.dorefreshwaste = True  # Have to recalculate waste


class Voter:
    def __init__(self, uid: str):
        self.uid = uid
        self.votelinks: List[VoteLink] = []  # Links to candidate in order of preference
        self._waste: float = 1
        self.dorefreshwaste = False  # Used to trigger recalculation of waste. Reduces computation
        self.doallocate = True

    def __repr__(self):
        return f"Voter({self.uid})"

    @property
    def waste(self) -> float:
        if self.dorefreshwaste:
            self.dorefreshwaste = False
            self._waste = 1 - sum(vl.weight for vl in self.votelinks)
        return self._waste

    def allocate_votes(self) -> None:
        self.doallocate = False

        # Collect all fixed weight and reset unfixed weight
        total = 1.0  # Total to allocate
        for vl in self.votelinks:
            if vl.status in [vl.PARTIAL, vl.FULL]:
                total -= vl.weight  # Removing fixed weight
            elif vl.weight > 0:
                vl.weight = 0
                vl.candidate.dorefreshvotes = True

        # Spread unfixed weight to first Active or Partial votelinks
        if total > 0.005:  # Due to floating point inaccuracy dont compare to 0
            for vl in self.votelinks:
                if vl.status in [vl.ACTIVE, vl.PARTIAL]:
                    vl.weight += total
                    total = 0
                    vl.candidate.dorefreshvotes = True
                    # New available support to previous winner
                    if vl.candidate.wonatquota > 0:
                        vl.candidate.doreduction = True
                    break

        self._waste = total


class VoteLink:
    EXCLUDED: Final = -2  # Permanent Lost support
    DEACTIVATED: Final = -1  # Temporary Lost support
    ACTIVE: Final = 0  # Active support
    PARTIAL: Final = 1  # Partial support
    FULL: Final = 2  # Full support

    def __init__(self, voter: Voter, candidate: Candidate):
        self.voter = voter
        self.candidate = candidate
        self.weight: float = 0

        self.voter.votelinks.append(self)
        self.candidate.votelinks.append(self)

        self.status = self.ACTIVE

    def __repr__(self):
        return f"VoteLink({self.voter}, {self.candidate}, {self.weight:.3f}, {self.status})"


class STVStatus:
    """ Result of each counting round """
    # Yield Levels. Used to see level of details
    INITIAL: Final = -1
    BEGIN: Final = 0
    END: Final = 1
    ROUND: Final = 2
    SUBROUND: Final = 3
    LOOP: Final = 4

    def __init__(self, yieldlevel: int = None):
        self.yieldlevel = yieldlevel  # Tells where in the algorithm the yield happened
        self.winner: Optional[Candidate] = None
        self.loser: Optional[Candidate] = None
        self.excluded_by_group: List[Candidate] = []
        self.reactivated: Optional[List[Candidate]] = None


class STV:
    """ Contains the whole voting system and does the counting """
    def __init__(self, usegroups: bool = False, reactivationmode: bool = False):
        # Static attributes
        self.usegroups = usegroups
        self.reactivationmode = reactivationmode

        # Input Attributes
        self.groups: Dict[str, Group] = {}
        self.candidates: Dict[str, Candidate] = {}
        self.voters: Dict[str, Voter] = {}

        # Variable Running Attributes
        self.totalseats = 0
        self.rounds = 0
        self.issubround = False
        self.subrounds = 0
        self.loopcount = 0
        self.allocationcount = 0
        self.reductioncount = 0

        self.winners: List[Candidate] = []
        self.active: List[Candidate] = []
        self.deactivated: List[Candidate] = []
        self.excluded: List[Candidate] = []

    # Setup Methods
    def add_group(self, name: str, seats: int) -> None:
        if name in self.groups:
            raise STVSetupException(f"Group {name} was already added")
        self.groups[name] = Group(name, seats)
        self.totalseats += seats

    def add_candidate(self, code: str, name: str, groupname: str) -> None:
        if code in self.candidates:
            raise STVSetupException(f"Candidate {code} was already added")
        try:
            group = self.groups[groupname]
        except KeyError:
            raise STVSetupException(f"Cannot find Group with name: {groupname}")
        self.candidates[code] = candidate = Candidate(code, name, group)
        self.active.append(candidate)  # Put all Candidates in the active list

    def add_voter(self, uid: str, candlist: List[str]) -> None:
        if not uid:
            raise STVSetupException("Cannot add Voter with empty code")
        if uid in self.voters:
            raise STVSetupException(f"Voter {uid} was already added")
        self.voters[uid] = newvoter = Voter(uid)
        addedcandidates = set()  # Used to check duplicate candidate code
        for ccode in candlist:
            try:
                if ccode not in addedcandidates:
                    VoteLink(newvoter, self.candidates[ccode])
                    addedcandidates.add(ccode)
                else:
                    print(f"Warning: Voter {uid} already specified candidate ({ccode}). Ignoring")
            except KeyError:
                print(f"Warning: Voter {uid} voted used an invalid Candidate Code ({ccode}). Ignoring")

    @property
    def quota(self) -> float:
        return len(self.voters) / self.totalseats

    @property
    def totalwaste(self) -> float:
        return float(len(self.voters)) - sum(c.votes for c in self.active + self.winners)

    def _sort_active(self) -> None:
        self.active.sort(key=lambda candidate: candidate.votes, reverse=True)

    def start(self) -> Generator:
        """ Advance to next Round. Either there will be a win, a loss or reactivation. Then do heavy counting """
        yield STVStatus(STVStatus.BEGIN)

        while True:
            # Manage Round counting
            if self.issubround:
                self.subrounds += 1
            else:
                self.rounds += 1
                self.subrounds = 1
            self.issubround = True
            self.loopcount = 0

            # Part 1: General Redistribution of votes
            loopstatus = STVStatus(STVStatus.LOOP)

            repeatmainloop = True
            while repeatmainloop:  # Main Loop
                repeatmainloop = False
                self.loopcount += 1

                for voter in self.voters.values():  # Allocation Loop
                    if voter.doallocate:
                        voter.allocate_votes()  # This can give surplus votes to candidates
                        self.allocationcount += 1
                if self.allocationcount > 0:
                    yield loopstatus
                    self.allocationcount = 0

                for winner in self.winners:  # Reduction Loop
                    if winner.doreduction:  # If candidate received surplus votes allocate_votes above
                        repeatmainloop = True  # Repeat Main loop
                        winner.reduce()  # Return surplus votes to voters and trigger doallocate
                        self.reductioncount += 1
                if self.reductioncount > 0:
                    yield loopstatus
                    self.reductioncount = 0

            self._sort_active()

            if self.rounds == 1 and self.subrounds == 1:  # Show pretty Initial Round for humans
                yield STVStatus(STVStatus.INITIAL)

            # Part 2: Decide
            decstatus = STVStatus()
            topcandidate = self.active[0]
            # Win. Either Quota is reached, or cannot lose a candidate because active list becomes too small
            if self.active[0].votes >= self.quota or len(self.winners) + len(self.active) == self.totalseats:
                # Register at which vote amount the winner won in case he won below the quota
                topcandidate.wonatquota = self.quota if topcandidate.votes > self.quota else topcandidate.votes
                # Status set to PARTIAL and let Candidate's Reduce function decide if FULL
                self._process_candidate(topcandidate, self.active, self.winners, VoteLink.PARTIAL, False)
                topcandidate.doreduction = True

                # Update status
                decstatus.winner = topcandidate

                # Update candidate's group
                wgroup = topcandidate.group
                wgroup.seatswon += 1

                # Process group constraints if groupquota is on
                if self.usegroups and wgroup.is_full:
                    for c in self.active + self.deactivated:
                        if c.group == wgroup:
                            fromlist = self.active if c in self.active else self.deactivated
                            self._process_candidate(c, fromlist, self.excluded, VoteLink.EXCLUDED, True)
                            decstatus.excluded_by_group.append(c)

                if len(self.winners) == self.totalseats:  # Finish and exit loop
                    decstatus.yieldlevel = decstatus.END
                    yield decstatus
                    return
                elif self.reactivationmode:  # If win and not finished and reactivationmode is on
                    decstatus.reactivated = self._reactivate()
                self.issubround = False

            else:  # Lose
                # Remove last active candidate
                roundloser = self.active[-1]
                self._process_candidate(roundloser, self.active, self.deactivated, VoteLink.DEACTIVATED, True)
                decstatus.loser = roundloser

            # If there are not enough active candidates, reactivate some.
            missingseats = self.totalseats - len(self.winners) - len(self.active)
            if missingseats > 0:
                # This can happen if reactivationmode is off and group quotas exclude too many candidates
                decstatus.reactivated = self._reactivate(missingseats)
                if len(decstatus.reactivated) != missingseats:
                    raise Exception('Reactivation failed in Round {}.{}'.format(self.rounds, self.subrounds))

            decstatus.yieldlevel = decstatus.SUBROUND if self.issubround else decstatus.ROUND
            yield decstatus

    @staticmethod
    def _process_candidate(candidate: Candidate,
                           fromlist: List[Candidate],
                           tolist: List[Candidate],
                           new_vl_status,
                           votersdoallocate
                           ) -> None:
        """ Transfer candidate and update its votelinks """
        fromlist.remove(candidate)
        tolist.append(candidate)

        for vl in candidate.votelinks:
            vl.status = new_vl_status
            vl.voter.doallocate = votersdoallocate

    def _reactivate(self, limit: Optional[int] = None) -> List[Candidate]:
        """ Reactivates deactivated Candidates """
        reactivated = []
        for c in self.deactivated[::-1]:
            self._process_candidate(c, self.deactivated, self.active, VoteLink.ACTIVE, True)
            reactivated.append(c)

            # If limit is set return the specified amount instead of all deactivated
            if limit is not None and len(reactivated) >= limit:
                break

        return reactivated
